package weka.classifiers.misc;

import java.text.NumberFormat;

import Jama.Matrix;

/**
 * 
 * @author sdanzige
 *
 * Generate a test set for the polynomial classifier
 */
class GenTestSet {
	private Matrix x; //The feature vectors
	private Matrix y; //The classification vectors
	
	public GenTestSet (){
				
	}
	
	/**
	 * x1, 2*x2, 3*x3, ... m*xm.  y is "1" for highest
	 * @param i //number of instances
	 * @param m //number of attributes
	 */
	public void algorithm1 (int i,int m) {
		x = new Matrix(i,m);
		y = new Matrix(i,m);
		
		int curMaxIdx = 0;
		double curMax = 0;
		for (int instance = 0; instance < i; instance++){
			for (int attribute = 0; attribute < m; attribute++){
				double curAttr = StrictMath.random()*2*m;
				if (curAttr > curMax) {
					curMax = curAttr;
					curMaxIdx = attribute;
				}
				x.set(instance,attribute,curAttr*(attribute+1));				
			}
			
			y.set(instance,curMaxIdx,1.00);
		}
		
	}
	
	/**
	 * x1, x2, x3....
	 * @param i
	 * @param m
	 */
	public void algorithm2 (int i,int m) {
		x = new Matrix(i,m);
		y = new Matrix(i,m);
		
		int curMaxIdx = 0;
		double curMax = 0;
		for (int instance = 0; instance < i; instance++){
			for (int attribute = 0; attribute < m; attribute++){
				double curAttr = StrictMath.random()*2*m;
				if (curAttr > curMax) {
					curMax = curAttr;
					curMaxIdx = attribute;
				}
				x.set(instance,attribute,curAttr);				
			}
			
			y.set(instance,curMaxIdx,1.00);
		}
		
	}
	
	public Matrix returnX(){
		return x;
	}
	
	public Matrix returnY(){
		return y;
	}
	
	public static void main (String args []) {
    	GenTestSet myGTS = new GenTestSet();
    	myGTS.algorithm1(30,5);
    	
    	myGTS.x.print(NumberFormat.getInstance(), 3);
		myGTS.y.print(NumberFormat.getInstance(), 3);
    		
	}
}