package weka.classifiers.misc;
import java.util.ArrayList;
import java.util.Date;

import Jama.CholeskyDecomposition;
import Jama.Matrix;
/**
 * 
 * @author sdanzige
 *
 *  Run the polynet classifier in java
 *  Comments
 */
class Polynet implements java.io.Serializable{
	Matrix w = null;

	private ArrayList errList;
	/**
	 * 
	 * @param pinv_type  Set to Cholesky to use Cholesky decomposition
	 * @param numEpoch   How many epochs to train for
	 */
	public Polynet (int numEpoch, String pinv_type, Matrix x, Matrix y) {		
		Matrix newY = null;
		
		
		//Calculate A = Xt * X and B = Xt * Y
		Matrix Am = (x.transpose()).times(x);
		Matrix B = (x.transpose()).times(y);
		errList = null;
		if (pinv_type.equalsIgnoreCase("Cholesky")) {
			CholeskyDecomposition A = new CholeskyDecomposition(Am);
			w = A.solve(B);
		} else {
			//Regular Pseudo-Inverse
			w = (Am.inverse()).times(B); 
		}
		
		newY = x.times(w);
		System.out.println("Pseudoinverse Accuracy = "+calcAccuracy(y,newY));
		
		//Iterated Learning
		for (int i=0; i<numEpoch; i++){
			Date t1 = new Date();

			errList = countWrong(y,newY);
			int [] errIdxs = new int[errList.size()];
			for (int j=0; j<errList.size(); j++){
				errIdxs[j]=((Integer)(errList.get(j))).intValue();				
			}
			
			Matrix wrongX = x.getMatrix(errIdxs,0,x.getColumnDimension()-1);
			Matrix wrongY = y.getMatrix(errIdxs,0,y.getColumnDimension()-1);
			
			Matrix addA = (wrongX.transpose()).times(wrongX);
			Matrix addB = (wrongX.transpose()).times(wrongY);
			
			Am = Am.plus(addA);
			B = B.plus(addB);
			
			Date t2 = new Date();
			
			
			if (pinv_type.equalsIgnoreCase("Cholesky")) {
				CholeskyDecomposition A = new CholeskyDecomposition(Am);
				w = A.solve(B);
			} else {
				//Regular Pseudo-Inverse
				w = (Am.inverse()).times(B); 
			}
			
					
			newY = x.times(w);		
			System.out.println(i+")\tTime: "+(t2.getTime()-t1.getTime())+"(ms)");
			System.out.println("\tPseudoinverse Accuracy = "+calcAccuracy(y,newY));
					
		}
		
		/*int [] ms= {5,10,15,20,30,40,60,80,100,200,300}; 
		
		for (int idx=0; idx < ms.length; idx++){
			timeTrials(5000,ms[idx]);
		}*/
		
	}
	
	
	private void timeTrials(int i, int m){

		System.out.println("\ni="+i);
		System.out.println("n,m="+m);
		GenTestSet myGTS = new GenTestSet();
		myGTS.algorithm1(i,m);
		Matrix trainX = myGTS.returnX();
		Matrix trainY = myGTS.returnY();
		//myGTS.algorithm1(10,100);
		//Matrix testX = myGTS.returnX();
		//Matrix testY = myGTS.returnY();
		
		//Solve the first time: LEAST SQUARES
		Date t1=null;
		Date t2=null;
		Matrix w = null;
		Matrix newY = null;
		
		if (true) {
			t1 = new Date();
			w = trainX.solve(trainY);
			t2 = new Date();
			System.out.println("Least Squares Solution: "+((t2.getTime()-t1.getTime())/1)+" ms");
		
			//See how accurate initial result is
			newY = trainX.times(w);
			System.out.println("Pseudoinverse Accuracy = "+calcAccuracy(trainY,newY));
		}
		
		//Calculate A = Xt * X and B = Xt * Y
		/*int m = trainX.getColumnDimension();
		int n = testX.getColumnDimension();
		Matrix A = new Matrix(m,m);
		Matrix B = new Matrix(m,n);*/
		t1 = new Date();
		Matrix Am = (trainX.transpose()).times(trainX);
		t2 = new Date();
		System.out.println("Am: "+((t2.getTime()-t1.getTime())/1)+" ms");

		t1 = new Date();
		Matrix B = (trainX.transpose()).times(trainY);
		t2 = new Date();
		System.out.println("B: "+((t2.getTime()-t1.getTime())/1)+" ms");

		t1 = new Date();
		CholeskyDecomposition A = new CholeskyDecomposition(Am);
		t2 = new Date();
		System.out.println("A: "+((t2.getTime()-t1.getTime())/1)+" ms");
		System.out.println("Is SPD? "+A.isSPD());
		
		t1 = new Date();
		w = A.solve(B);
		t2 = new Date();
		System.out.println("Cholesky W: "+((t2.getTime()-t1.getTime())/1)+" ms");
		newY = trainX.times(w);
		System.out.println("Pseudoinverse Accuracy = "+calcAccuracy(trainY,newY));

		t1 = new Date();
		w = (Am.inverse()).times(B);
		t2 = new Date();
		System.out.println("Inverse W: "+((t2.getTime()-t1.getTime())/1)+" ms");
		newY = trainX.times(w);
		System.out.println("Pseudoinverse Accuracy = "+calcAccuracy(trainY,newY));
	}
	
	/**
	 * Compare location of max value of row in m1 and m2
	 * If they're the same, then add one true, else 1 false
	 * @param m1
	 * @param m2
	 */
	private double calcAccuracy(Matrix m1, Matrix m2){
		int correct=0;
		int wrong=0;
		
		errList = countWrong(m1,m2);
		
		double w = errList.size();
		double c = m1.getRowDimension() - w;
		
		double acc = c / (c+w);
		return acc;		
	}
	
	/**
	 * Determine which rows are misclassified
	 * @param mtrx1
	 * @param mtrx2
	 * @return
	 */
	private ArrayList countWrong(Matrix m1, Matrix m2){
		ArrayList errList = new ArrayList();

		double [][] mtrx1 = m1.getArray();
		double [][] mtrx2 = m2.getArray();
		int numInst = m1.getRowDimension();
		
		for (int i=0; i<numInst; i++){
			double [] tm1 = mtrx1[i];
			double [] tm2 = mtrx2[i];
			//int idx1 = idxMaxVal(tm1);
			//int idx2 = idxMaxVal(tm2);
			int idx1 = idxNearest1(tm1);
			int idx2 = idxNearest1(tm2);
			
			if (idx1 == idx2){
				//correct++;
			} else {
				errList.add(new Integer(i));
			}
		}
		
		return errList;
	}
	
	/**
	 * Return the index of the max value in the array
	 * @return
	 */
	private int idxMaxVal(double vect []){
		double maxVal=0;
		int idx=0;
		for (int i=0; i<vect.length; i++){
			if (vect[i] > maxVal){
				maxVal = vect[i];
				idx = i;
			}
		}
		
		return idx;
	}
	
	private int idxNearest1(double vect []){
		double curClosest=Double.MAX_VALUE;
		int idx = 0;
		for (int i=0; i<vect.length; i++){
			double test = Math.abs(vect[i]-1);
			if (test<curClosest){
				curClosest = test;
				idx = i;
			}
		}
		return idx;
	}
	
	/**
	 *  Return the classification index of a new instance
	 *  @param newX
	 *  @return
	 */
	public int classify(Matrix newX){
		Matrix predY = newX.times(w);
		double [][] mY = predY.getArray();
		int idx = idxNearest1(mY[0]);
		return idx;
	}
	
	public static void main (String args []) {
		GenTestSet myGTS = new GenTestSet();
		myGTS.algorithm1(5000,30);
		//myGTS.algorithm2(5000,30);
		Matrix trainX = myGTS.returnX();
		Matrix trainY = myGTS.returnY();

		//myGTS.algorithm1(10,100);
		//Matrix testX = myGTS.returnX();
		//Matrix testY = myGTS.returnY(); 		

		//Polynet myPN = new Polynet(100,"Cholesky",trainX,trainY);
		Polynet myPN = new Polynet(100,"no",trainX,trainY);  
		
	}

}